#include "llvm/Pass.h"
#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/Instructions.h"
#include "llvm/PassManager.h"
#include "llvm/Support/SourceMgr.h"
#include "llvm/Support/raw_ostream.h"

#include "llvm/Analysis/MemoryDependenceAnalysis.h"
#include "llvm/Analysis/Passes.h"

#include <vector>
#include <list>
#include <map>
#include <cassert>

using namespace llvm;

static void collectLoadsStores (Function * f,
                                std::vector<LoadInst*>& loads,
                                std::vector<StoreInst*>& stores) {
  assert(f && "NULL parameter function");
  for (auto &b : *f) {
    for (auto &i : b) {
      // i.dump();
      LoadInst * l_inst = dyn_cast<LoadInst>(&i);
      if (l_inst) {
        loads.push_back(l_inst);
        continue;
      }

      StoreInst * s_inst = dyn_cast<StoreInst>(&i);
      if (s_inst)
        stores.push_back(s_inst);
    }
  }
}

namespace llvm {

  class playdep_kernel : public ModulePass {
  public:
    static char ID;

    playdep_kernel() : ModulePass(ID) {}
    
    void getAnalysisUsage(AnalysisUsage &AU) const {
      AU.addRequired<AliasAnalysis>();
      AU.addRequired<MemoryDependenceAnalysis>();
    }

    virtual bool runOnModule(Module &m) {

      LLVMContext &Context = m.getContext();

      for (auto &f : m) {
        if (f.isDeclaration()) continue;

        errs() << "function " << f.getName() << "\n";
	if (f.getName().equals("main")) continue;
        
        std::vector<LoadInst*> loads;
        std::vector<StoreInst*> stores;
        collectLoadsStores(&f, loads, stores);
        std::map<int, std::vector<Value*>> layer_indirection;

        if (stores.size() < 1) continue;
        // f.viewCFG();
	StoreInst *s_inst = stores[1];

        errs() << "for this instruction: ";
        s_inst->dump();

        errs() << "deps are: \n";

        MemoryDependenceAnalysis &mda = getAnalysis<MemoryDependenceAnalysis>(f);
        
        MemDepResult mdr = mda.getDependency (s_inst);
        while (mdr.getInst()) {
          mdr.getInst()->dump();
          mdr = mda.getDependency (mdr.getInst());
        }
      }

      return false;
    }
    
  };

  class playdep : public ModulePass {
  public:
    static char ID;

    playdep() : ModulePass(ID) {}

    virtual bool runOnModule(Module &m) {
      PassManager PM;
      PM.add(createBasicAliasAnalysisPass());
      PM.add(new playdep_kernel());
      PM.run(m);
    }

  };

  char playdep_kernel::ID = 0;
  char playdep::ID = 0;
  static RegisterPass<playdep>
  X("playdep", "playdep Pass", false, false);

  class PassRegistry;
  void initializeplaydepPass(PassRegistry &);
}

INITIALIZE_PASS_BEGIN(playdep, "playdep",
                      "playdep", false, false)
INITIALIZE_PASS_DEPENDENCY(MemoryDependenceAnalysis)
INITIALIZE_AG_DEPENDENCY(AliasAnalysis)
INITIALIZE_PASS_END(playdep, "playdep",
                    "playdep", false, false)
